package ch.fritteli.gombaila.domain.generator;

import ch.fritteli.gombaila.domain.common.NodeBinExprAdd;
import ch.fritteli.gombaila.domain.common.NodeBinExprDiv;
import ch.fritteli.gombaila.domain.common.NodeBinExprExp;
import ch.fritteli.gombaila.domain.common.NodeBinExprMinus;
import ch.fritteli.gombaila.domain.common.NodeBinExprMod;
import ch.fritteli.gombaila.domain.common.NodeBinExprMult;
import ch.fritteli.gombaila.domain.common.NodeExpr;
import ch.fritteli.gombaila.domain.common.NodeExprIdent;
import ch.fritteli.gombaila.domain.common.NodeExprIntLit;
import io.vavr.control.Option;
import org.jetbrains.annotations.NotNull;

final class ExprVisitor {
    @NotNull
    private final Generator generator;

    ExprVisitor(@NotNull final Generator generator) {
        this.generator = generator;
    }

    void visit(@NotNull final NodeExpr expr) {
        switch (expr) {
            case final NodeExprIdent exprIdent -> visit(exprIdent);
            case final NodeExprIntLit exprIntLit -> visit(exprIntLit);
            case final NodeBinExprAdd exprAdd -> visit(exprAdd);
            case final NodeBinExprMinus exprMinus -> visit(exprMinus);
            case final NodeBinExprMult exprMult -> visit(exprMult);
            case final NodeBinExprDiv exprDiv -> visit(exprDiv);
            case final NodeBinExprMod exprMod -> visit(exprMod);
            case final NodeBinExprExp exprExp -> visit(exprExp);
        }
    }

    private void visit(@NotNull final NodeExprIdent expr) {
        final String varName = (String) expr.ident().value().get();
        final Option<Variable> variable = this.generator.vars.find(var -> var.name().equals(varName));

        if (variable.isEmpty()) {
            throw new GeneratorException("Undeclared identifier: %s".formatted(varName), expr);
        }

        generator.push("qword [rsp+%d]".formatted(8 * (this.generator.stackSize - variable.get().stackLocation() - 1)));
    }

    private void visit(@NotNull final NodeExprIntLit expr) {
        this.generator.printer.commentedLine("store the value in rax", "mov", "rax, %s".formatted(expr.value().value().get()));
        generator.push("rax");
    }

    private void visit(@NotNull final NodeBinExprAdd expr) {
        generator.generateExpr(expr.lhs());
        generator.generateExpr(expr.rhs());
        generator.pop("rbx");
        generator.pop("rax");
        generator.printer.commentedLine("add rbx to rax", "add", "rax, rbx");
        generator.push("rax");
    }

    private void visit(@NotNull final NodeBinExprMinus expr) {
        generator.generateExpr(expr.lhs());
        generator.generateExpr(expr.rhs());
        generator.pop("rbx");
        generator.pop("rax");
        generator.printer.commentedLine("subtract rbx from rax", "sub", "rax, rbx");
        generator.push("rax");
    }

    private void visit(@NotNull final NodeBinExprMult expr) {
        generator.generateExpr(expr.lhs());
        generator.generateExpr(expr.rhs());
        generator.pop("rbx");
        generator.pop("rax");
        generator.printer.commentedLine("multiply rax by rbx", "mul", "rbx");
        generator.push("rax");
    }

    private void visit(@NotNull final NodeBinExprDiv expr) {
        generator.generateExpr(expr.lhs());
        generator.generateExpr(expr.rhs());
        generator.pop("rbx");
        generator.pop("rax");
        generator.printer.commentedLine("divide rax by rbx", "div", "rbx");
        generator.push("rax");
    }

    private void visit(@NotNull final NodeBinExprMod expr) {
        generator.generateExpr(expr.lhs());
        generator.generateExpr(expr.rhs());
        generator.pop("rbx");
        generator.pop("rax");
        generator.printer.commentedLine("divide rax by rbx. The remainder will be in rdx.", "div", "rbx");
        generator.push("rdx");
    }

    private void visit(@NotNull final NodeBinExprExp expr) {
        throw new GeneratorException("Not yet implemented!", expr);
    }
}
