package ch.fritteli.gombaila.domain.generator;

import ch.fritteli.gombaila.domain.common.NodeExpr;
import ch.fritteli.gombaila.domain.common.NodeProg;
import ch.fritteli.gombaila.domain.common.NodeStmt;
import ch.fritteli.gombaila.domain.printer.Printer;
import io.vavr.collection.List;
import io.vavr.collection.Seq;
import org.jetbrains.annotations.NotNull;

import java.time.ZonedDateTime;
import java.util.Stack;

public class Generator {
    @NotNull
    final Printer printer = new Printer();
    @NotNull
    final Stack<Integer> scopes = new Stack<>();
    @NotNull
    private final NodeProg nodeProg;
    @NotNull
    private final ExprVisitor exprVisitor;
    @NotNull
    private final StmtVisitor stmtVisitor;
    @NotNull
    Seq<Variable> vars = List.empty();
    int stackSize = 0;

    public Generator(@NotNull final NodeProg nodeProg) {
        this.nodeProg = nodeProg;
        this.exprVisitor = new ExprVisitor(this);
        this.stmtVisitor = new StmtVisitor(this);
    }

    @NotNull
    public String generate() {
        this.generateHeader();

        for (final NodeStmt stmt : this.nodeProg.stmts()) {
            this.generateStmt(stmt);
        }

        this.generateFooter();

        return this.printer.toString();
    }

    private void generateHeader() {
        this.printer.comment("Program generated by Gombaila on %1$tF, %1$tT".formatted(ZonedDateTime.now()));
        this.printer.section(".data");
        this.printer.line("EXIT_SUCCESS", "equ", 0);
        this.printer.line("SYSCALL_exit", "equ", 60);
        this.printer.blank();
        this.printer.section(".text");
        this.printer.rawLine("global _start");
        this.printer.blank();
        this.printer.label("_start");
    }

    private void generateFooter() {
        this.printer.commentedLine("Prepare for successful exit", "mov", "rdi, EXIT_SUCCESS");
        this.printer.commentedLine("Jump to the default exit routine", "jmp", "exit");
        this.printer.blank();
        this.printer.label("exit");
        this.printer.commentedLine("Prepare for the EXIT syscall", "mov", "rax, SYSCALL_exit");
        this.printer.line("syscall");
    }

    void generateStmt(@NotNull final NodeStmt stmt) {
        this.stmtVisitor.visit(stmt);
    }

    void generateExpr(@NotNull final NodeExpr expr) {
        this.exprVisitor.visit(expr);
    }

    void push(@NotNull final String reg) {
        this.printer.commentedLine("push %s onto top of the stack".formatted(reg), "push", reg);
        this.stackSize++;
    }

    void pop(@NotNull final String reg) {
        this.printer.commentedLine("pop top of the stack into %s".formatted(reg), "pop", reg);
        this.stackSize--;
    }

    void beginScope() {
        this.scopes.push(this.vars.size());
    }

    void endScope() {
        final int popCount = this.vars.size() - this.scopes.pop();
        if (popCount > 0) {
            this.printer.commentedLine(
                    "Reset the stack pointer to discard out-of-scope variables",
                    "add", "rsp, %d".formatted(popCount * 8));
            this.stackSize -= popCount;
        }
        this.vars = this.vars.dropRight(popCount);
    }
}
