package ch.fritteli.gombaila.domain.generator;

import ch.fritteli.gombaila.domain.common.NodeStmt;
import ch.fritteli.gombaila.domain.common.NodeStmtAssign;
import ch.fritteli.gombaila.domain.common.NodeStmtExit;
import ch.fritteli.gombaila.domain.common.NodeStmtIf;
import ch.fritteli.gombaila.domain.common.NodeStmtIfElse;
import ch.fritteli.gombaila.domain.common.NodeStmtLet;
import ch.fritteli.gombaila.domain.common.NodeStmtPrint;
import ch.fritteli.gombaila.domain.common.NodeStmtScope;
import io.vavr.control.Option;
import org.jetbrains.annotations.NotNull;

final class StmtVisitor {
    @NotNull
    private final Generator generator;
    private int labelCounter = 0;

    StmtVisitor(@NotNull final Generator generator) {
        this.generator = generator;
    }

    void visit(@NotNull final NodeStmt stmt) {
        switch (stmt) {
            case final NodeStmtLet stmtLet -> visit(stmtLet);
            case final NodeStmtAssign stmtAssign -> visit(stmtAssign);
            case final NodeStmtExit stmtExit -> visit(stmtExit);
            case final NodeStmtPrint stmtPrint -> visit(stmtPrint);
            case final NodeStmtScope stmtScope -> visit(stmtScope);
            case final NodeStmtIf stmtIf -> visit(stmtIf);
            case final NodeStmtIfElse stmtIfElse -> visit(stmtIfElse);
        }
    }

    private void visit(@NotNull final NodeStmtLet stmt) {
        final String varName = (String) stmt.stmtAssign().ident().value().get();
        if (this.generator.vars.exists(var -> var.name().equals(varName))) {
            throw new GeneratorException("Identifier already used: %s".formatted(varName), stmt);
        }

        this.generator.vars = this.generator.vars.append(new Variable(varName, generator.stackSize));

        this.generator.generateExpr(stmt.stmtAssign().expr());
    }

    private void visit(@NotNull final NodeStmtAssign stmt) {
        final String varName = (String) stmt.ident().value().get();
        final Option<Variable> variable = this.generator.vars.find(var -> var.name().equals(varName));
        if (variable.isEmpty()) {
            throw new GeneratorException("Undeclared identifier: %s".formatted(varName), stmt);
        }

        final int stackLocation = variable.get().stackLocation();
        final int targetLocation = 8 * (this.generator.stackSize - stackLocation - 1);
        this.generator.generateExpr(stmt.expr());
        this.generator.pop("rbx");
        this.generator.printer.commentedLine(
                "Overwrite the old value of %s in the stack",
                "mov",
                "[rsp+%d], rbx".formatted(targetLocation)
        );
    }

    private void visit(@NotNull final NodeStmtExit stmt) {
        this.generator.generateExpr(stmt.expr());
        this.generator.pop("rdi");
        this.generator.printer.commentedLine("Jump to the default exit routine", "jmp", "exit");
    }

    private void visit(@NotNull final NodeStmtPrint stmt) {
//            generateExpr(stmt.nodeExpr());
//            // rdx: length of output
//            line("mov rdx, 8");
//            // rsi: output
//            pop("rsi");
//            // rdi: fd-value (STDOUT=1)
//            line("mov rdi, 1");
//            // rax: syscall (SYS_WRITE=1)
//            line("mov rax, 1");
//            // syscall
//            line("syscall");
        throw new GeneratorException("Not implemented yet", stmt);
    }

    private void visit(@NotNull final NodeStmtScope stmt) {
        this.generator.beginScope();
        stmt.stmts().forEach(this.generator::generateStmt);
        this.generator.endScope();
    }

    private void visit(@NotNull final NodeStmtIf stmt) {
        final String label = this.createNextLabel();
        this.generator.generateExpr(stmt.expr());
        this.generator.pop("rax");
        this.generator.printer.commentedLine("test the condition",
                "test", "rax, rax");
        this.generator.printer.commentedLine(
                "skip to %s if condition is not met".formatted(label),
                "jz", label
        );
        this.visit(stmt.ifScope());
        this.generator.printer.label(label);
    }

    private void visit(@NotNull final NodeStmtIfElse stmt) {
        final String afterIfLabel = this.createNextLabel();
        final String afterElseLabel = this.createNextLabel();
        this.generator.generateExpr(stmt.expr());
        this.generator.pop("rax");
        this.generator.printer.commentedLine("test the condition",
                "test", "rax, rax");
        this.generator.printer.commentedLine(
                "skip to %s (else) if condition is not met".formatted(afterIfLabel),
                "jz", afterIfLabel
        );
        this.visit(stmt.ifScope());
        this.generator.printer.commentedLine(
                "skip to %s to skip the else".formatted(afterElseLabel),
                "jmp",
                afterElseLabel
        );
        this.generator.printer.label(afterIfLabel);
        this.visit(stmt.elseScope());
        this.generator.printer.label(afterElseLabel);
    }

    @NotNull
    private String createNextLabel() {
        return "label" + (this.labelCounter++);
    }
}
