package ch.fritteli.gombaila.domain.lexer;

import ch.fritteli.gombaila.domain.common.Token;
import ch.fritteli.gombaila.domain.common.TokenType;
import io.vavr.Predicates;
import io.vavr.collection.Stream;
import io.vavr.control.Option;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

public class Lexer {
    @NotNull
    private final StringWalker content;
    @NotNull
    private Stream<Token> tokens = Stream.empty();

    public Lexer(@NotNull final String input) {
        this.content = new StringWalker(input);
    }

    @NonNls
    public Stream<Token> lex() {
        Option<Character> next;
        while ((next = this.content.peekNext()).isDefined()) {
            final char c = next.get();
            final int line = this.content.line();
            final int column = this.content.column();
            if (Character.isAlphabetic(c) || c == '_') {
                this.handleAlphabeticOrUnderscore(line, column);
            } else if (Character.isDigit(c)) {
                this.handleDigit(line, column);
            } else if (Character.isWhitespace(c)) {
                this.handleWhitespace(line, column);
            } else if (c == ';') {
                this.handleSimple(TokenType.SEMI, line, column);
            } else if (c == '(') {
                this.handleSimple(TokenType.OPEN_PAREN, line, column);
            } else if (c == ')') {
                this.handleSimple(TokenType.CLOSE_PAREN, line, column);
            } else if (c == '{') {
                this.handleSimple(TokenType.OPEN_CURLY, line, column);
            } else if (c == '}') {
                this.handleSimple(TokenType.CLOSE_CURLY, line, column);
            } else if (c == '=') {
                this.handleSimple(TokenType.EQUALS, line, column);
            } else if (c == '+') {
                this.handleSimple(TokenType.PLUS, line, column);
            } else if (c == '-') {
                this.handleSimple(TokenType.MINUS, line, column);
            } else if (c == '*') {
                this.handleSimple(TokenType.MULT, line, column);
            } else if (c == '/') {
                this.handleForwardSlash(line, column);
            } else if (c == '%') {
                this.handleSimple(TokenType.MOD, line, column);
            } else if (c == '^') {
                this.handleSimple(TokenType.EXP, line, column);
            } else {
                throw this.error(c, line, column);
            }
        }
        return this.tokens;
    }

    private LexerException error(final char c, final int line, final int column) {
        return new LexerException("Error parsing input: Unexpected character '%c'.".formatted(c), line, column);
    }

    private void handleAlphabeticOrUnderscore(final int line, final int column) {
        final StringBuilder s = new StringBuilder();
        while (this.content.peekNext().exists(Predicates.<Character>anyOf(
                Character::isAlphabetic,
                Character::isDigit,
                c -> c == '_'
        ))) {
            s.append(this.content.next());
        }
        switch (s.toString()) {
            case "exit" -> this.appendToken(new Token(TokenType.EXIT, line, column));
            case "let" -> this.appendToken(new Token(TokenType.LET, line, column));
            case "print" -> this.appendToken(new Token(TokenType.PRINT, line, column));
            case "if" -> this.appendToken(new Token(TokenType.IF, line, column));
            case "else" -> this.appendToken(new Token(TokenType.ELSE, line, column));
            case final String value ->
                    this.appendToken(new Token(TokenType.IDENTIFIER, Option.of(value), line, column));
        }
    }

    private void handleDigit(final int line, final int column) {
        final StringBuilder s = new StringBuilder();
        while (this.content.peekNext().exists(Character::isDigit)) {
            s.append(this.content.next());
        }
        this.appendToken(new Token(TokenType.INT_LIT, Option.of(Long.parseLong(s.toString())), line, column));
    }

    private void handleWhitespace(final int line, final int column) {
        final StringBuilder s = new StringBuilder();
        while (this.content.peekNext().exists(Character::isWhitespace)) {
            s.append(this.content.next());
        }
        this.appendToken(new Token(TokenType.WHITESPACE, Option.of(s.toString()), line, column));
    }

    private void handleSimple(@NotNull final TokenType tokenType, final int line, final int column) {
        this.content.next();
        this.appendToken(new Token(tokenType, line, column));
    }

    private void handleForwardSlash(final int line, final int column) {
        this.content.next();
        if (this.content.peekNext().exists(n -> n == '/')) {
            this.content.next();
            this.handleLineComment();
        }
        if (this.content.peekNext().exists(n -> n == '*')) {
            this.content.next();
            this.handleBlockComment();
        } else {
            this.tokens.append(new Token(TokenType.DIV, line, column));
        }
    }

    private void handleLineComment() {
        while (this.content.peekNext().exists(next -> next != '\n')) {
            this.content.next();
        }
        if (this.content.peekNext().exists(next -> next == '\n')) {
            this.content.next();
        }
    }

    private void handleBlockComment() {
        // ...*/
        while (this.content.peekNext().exists(next -> next != '*')) {
            this.content.next();
        }
        if (this.content.peekNext().exists(n -> n == '*')) {
            this.content.next();
            if (this.content.peekNext().exists(n -> n == '/')) {
                this.content.next();
                return;
            }
            handleBlockComment();
        } else {
            return;
        }
    }

    private void appendToken(@NotNull final Token token) {
        this.tokens = this.tokens.append(token);
    }
}
